module AMR exposing
    ( Config
    , element, document, application
    , Program
    )

{-| This module helps you set up an Elm program with automatic message-replay.

I suggest that you setup your `Main` module to make it easy to toggle between this development-version of your program and the one for production. Maybe using a comment, like this:

    main =
    AMR.application amrConfig
        -- Browser.application
        { ...
        }

# Configuration

@docs Config

# Programs

This API mirrors [elm/browser](https://package.elm-lang.org/packages/elm/browser/latest).

@docs element, document, application

@docs Program

-}

import Browser
import Browser.Navigation
import DenseList exposing (DenseList)
import Html exposing (Html)
import Json.Decode
import Json.Encode
import Url exposing (Url)


{-| A program with automatic message-replay enabled.
-}
type alias Program flags model msg =
    Platform.Program flags (Model model) (Msg msg)


{-| This is the configuration required for automatic message-replay.

Let's say your program is initialized in HTML like this:

    <script>
        const app = Elm.Main.init({
            node: document.body,
            flags: {
                amr : localStorage.getItem("app-amr")
            }
        })

        app.ports.toCache.subscribe(output =>
            localStorage.setItem("app-amr", output)
        )
    </script>

The configuration for that script would be like this:

    port toCache : Maybe String -> Cmd msg

    fromCache =
        .amr

`encodeMsg` and `msgDecoder` depends how messages are defined. You can find an example [here](https://github.com/opvasger/amr/blob/master/example/src/Counter.elm#L21-L48).

-}
type alias Config flags msg =
    { encodeMsg : msg -> Json.Encode.Value
            , msgDecoder : Json.Decode.Decoder msg
            , fromCache : flags -> Maybe String
            , toCache : String -> Cmd (Msg msg)
            }


{-| Create an HTML element managed by Elm. More about that [here](https://package.elm-lang.org/packages/elm/browser/1.0.1/Browser#element).
-}
element :
    Config flags msg
    ->
        { init : flags -> ( model, Cmd msg )
        , view : model -> Html msg
        , update : msg -> model -> ( model, Cmd msg )
        , subscriptions : model -> Sub msg
        }
    -> Program flags model msg
element config app =
    Browser.element
        { update = update app.update config.toCache config.encodeMsg
        , subscriptions = Sub.map UpdateApp << app.subscriptions << .model
        , view = Html.map UpdateApp << app.view << .model
        , init =
            \flags ->
                init app.update
                    config.msgDecoder
                    (config.fromCache flags)
                    (app.init flags)
        }


{-| Create an HTML document managed by Elm. More about that [here](https://package.elm-lang.org/packages/elm/browser/1.0.1/Browser#document).
-}
document :
    Config flags msg
    ->
        { init : flags -> ( model, Cmd msg )
        , subscriptions : model -> Sub msg
        , update : msg -> model -> ( model, Cmd msg )
        , view : model -> Browser.Document msg
        }
    -> Program flags model msg
document config app =
    Browser.document
        { update = update app.update config.toCache config.encodeMsg
        , subscriptions = Sub.map UpdateApp << app.subscriptions << .model
        , view = mapDocument UpdateApp << app.view << .model
        , init =
            \flags ->
                init app.update
                    config.msgDecoder
                    (config.fromCache flags)
                    (app.init flags)
        }


{-| Create an application that manages Url changes. More about that [here](https://package.elm-lang.org/packages/elm/browser/1.0.1/Browser#application).
-}
application :
    Config flags msg
    ->
        { init : flags -> Url -> Browser.Navigation.Key -> ( model, Cmd msg )
        , update : msg -> model -> ( model, Cmd msg )
        , subscriptions : model -> Sub msg
        , view : model -> Browser.Document msg
        , onUrlChange : Url -> msg
        , onUrlRequest : Browser.UrlRequest -> msg
        }
    -> Program flags model msg
application config app =
    Browser.application
        { update = update app.update config.toCache config.encodeMsg
        , subscriptions = Sub.map UpdateApp << app.subscriptions << .model
        , view = mapDocument UpdateApp << app.view << .model
        , onUrlChange = UpdateApp << app.onUrlChange
        , onUrlRequest = UpdateApp << app.onUrlRequest
        , init =
            \flags url key ->
                init app.update
                    config.msgDecoder
                    (config.fromCache flags)
                    (app.init flags url key)
        }



-- Program


type alias Model model =
    { model : model
    , msgs : DenseList String
    }


type Msg msg
    = UpdateApp msg


init :
    (msg -> model -> ( model, Cmd msg ))
    -> Json.Decode.Decoder msg
    -> Maybe String
    -> ( model, Cmd msg )
    -> ( Model model, Cmd (Msg msg) )
init updateApp msgDecoder fromCache ( model, appCmd ) =
    case Maybe.map (Json.Decode.decodeString (decoder (withoutCmd updateApp) msgDecoder model)) fromCache of
        Just (Ok cacheModel) ->
            ( cacheModel
            , Cmd.none
            )

        _ ->
            ( { model = model
              , msgs = DenseList.empty
              }
            , Cmd.map UpdateApp appCmd
            )


update :
    (msg -> model -> ( model, Cmd msg ))
    -> (String -> Cmd (Msg msg))
    -> (msg -> Json.Encode.Value)
    -> Msg msg
    -> Model model
    -> ( Model model, Cmd (Msg msg) )
update updateApp toCache encodeMsg msg model =
    case msg of
        UpdateApp updateMsg ->
            let
                ( updateModel, appCmd ) =
                    updateApp updateMsg model.model

                updatedModel =
                    { model
                        | model = updateModel
                        , msgs = DenseList.insert (Json.Encode.encode 0 (encodeMsg updateMsg)) model.msgs
                    }
            in
            ( updatedModel
            , Cmd.batch
                [ Cmd.map UpdateApp appCmd
                , toCache (Json.Encode.encode 0 (encode updatedModel))
                ]
            )


encode : Model model -> Json.Encode.Value
encode model =
    DenseList.encodeReversed Json.Encode.string model.msgs


decoder :
    (msg -> model -> model)
    -> Json.Decode.Decoder msg
    -> model
    -> Json.Decode.Decoder (Model model)
decoder replayUpdate msgDecoder model =
    Json.Decode.map
        (\( msgs, updatedModel ) ->
            { model = updatedModel
            , msgs = msgs
            }
        )
        (DenseList.foldLeftUntilErrorDecoder
            Json.Decode.string
            (Json.Decode.decodeString msgDecoder)
            replayUpdate
            model
        )



-- Helpers


withoutCmd : (msg -> model -> ( model, Cmd msg )) -> msg -> model -> model
withoutCmd updateApp msg model =
    Tuple.first (updateApp msg model)


mapDocument : (a -> b) -> Browser.Document a -> Browser.Document b
mapDocument msg { title, body } =
    { title = title
    , body = List.map (Html.map msg) body
    }
